/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*iso_ionize_recombine find state specific creation and destruction rates for iso sequences */
/*ChargeTransferUpdate update rate of ct ionization and recombination for H atoms */
#include "cddefines.h"
#include "ionbal.h"
#include "atmdat.h"
#include "dense.h"
#include "grainvar.h"
#include "hmi.h"
#include "mole.h"
#include "secondaries.h"
#include "iso.h"
/*lint -e778 constant expression evaluates to zero - in HMRATE macro */
/*ChargeTransferUpdate update rate of ct ionization and recombination for H atoms */
static void ChargeTransferUpdate(void)
{
	long int ion;
	long int nelem;
	/* find total rate for charge transfer ionization of hydrogen,
	 * recombination for other element is ionization of hydrogen */
	atmdat.HCharExcIonTotal = 0.;
	for( nelem=ipHELIUM; nelem<LIMELM; ++nelem)
	{
		/* ion on the C scale, 0 is atom, so goes up to nelem+1,
		 * for helium nelem=1, ions go from 1 to 2 */
		for( ion=1; ion<=nelem+1; ++ion )
		{
			double one;
			/* we intentionally skip CT with O+ since this is in hmole */
			/* >>chng 05 aug 17, add logic on !ionbal.lgHO_ct_chem
			 * this is flag that is true when H O charge transfer is done in
			 * chemistry, not here.  set false with set HO charge transfer ionization
			 * command, so we do it here - introduced by NA for the Leiden hacks */
			if( (nelem == ipOXYGEN) && (ion == 1) && ionbal.lgHO_ct_chem ) 
				continue;
			/* charge transfer ionization of H, recombination for other species */
			one = atmdat.HCharExcRecTo[nelem][ion-1]*dense.xIonDense[nelem][ion];
			atmdat.HCharExcIonTotal += one;
		}
	}

	/* >>chng 01 may 07,  add this in */
	/* charge transfer recombination of hydrogen,
	 * which is ionization of the heavy element */
	atmdat.HCharExcRecTotal = 0.;
	for( nelem=ipHELIUM; nelem<LIMELM; ++nelem)
	{
		/* this is ion on the abundances scale, 1 is atom, so goes up to nelem+1,
		 * for helium nelem=1, ion must go up to 3 */
		for( ion=0; ion<=nelem; ++ion )
		{
			/* option to skip Oo => O+ */
			/* >>chng 05 aug 17, add logic on !ionbal.lgHO_ct_chem
			 * this is flag that is true when H O charge transfer is done in
			 * chemistry, not here.  set false with set HO charge transfer ionization
			 * command, so we do it here - introduced by NA for the Leiden hacks */
			if( (nelem == ipOXYGEN) && (ion == 0) && ionbal.lgHO_ct_chem ) 
				continue;
			/* charge transfer ionization of H, recombination for other species */
			atmdat.HCharExcRecTotal += 
				atmdat.HCharExcIonOf[nelem][ion]*dense.xIonDense[nelem][ion];
		}
	}

	/* >>logic checked 02 may 02, think it's right */
	/* add on charge transfer ionization of helium,
	 * recombination for other element is ionization of helium */
	atmdat.HeCharExcIonTotal = 0.;
	/* loop up from Li, the next heavier element */
	for( nelem=ipLITHIUM; nelem<LIMELM; ++nelem)
	{
		double hold_one = 0.;
		/* check that array bounds not exceeded */
		/* ion on the C scale, 0 is atom, so goes up to nelem+1,
		 * for helium nelem=1, ions go from 1 to 2 */
		for( ion=1; ion<=nelem+1; ++ion )
		{
			/* charge transfer ionization of He, recombination for other species */
			hold_one = atmdat.HeCharExcRecTo[nelem][ion-1]*dense.xIonDense[nelem][ion];
			atmdat.HeCharExcIonTotal += hold_one;
		}
	}

	/* >>chng 04 jul 02,
	 * add on charge transfer reactions of He-H */
	atmdat.HeCharExcIonTotal += atmdat.HCharExcIonOf[ipHELIUM][0]*dense.xIonDense[ipHYDROGEN][1];

	/* charge transfer recombination of helium,
	 * which is ionization of the heavy element */
	atmdat.HeCharExcRecTotal = 0.;
	for( nelem=ipLITHIUM; nelem<LIMELM; ++nelem)
	{
		/* this is ion on the physics scale, 1 is atom, so goes up to nelem+1,
		 * for helium nelem=1, ion must go up to 3 */
		for( ion=0; ion<=nelem; ++ion )
		{
			/* charge transfer recombination of He, ionization for other species */
			atmdat.HeCharExcRecTotal += 
				atmdat.HeCharExcIonOf[nelem][ion]*dense.xIonDense[nelem][ion];
		}
	}
	/* >>chng 04 jul 02
	 * Add on charge transfer reactions of He+ +H0 -> He0 + H+ */
	atmdat.HeCharExcRecTotal += atmdat.HCharExcRecTo[ipHELIUM][0]*dense.xIonDense[ipHYDROGEN][0];

	return;
}

/*iso_ionize_recombine find state specific creation and destruction rates for iso sequences */
void iso_ionize_recombine(
	/* iso sequence, hydrogen or helium for now */
	long ipISO , 
	/* the chemical element, 0 for hydrogen */
	long int nelem )
{
	long int level,
		ion;
	double Recom3Body,
		sum;

	DEBUG_ENTRY( "iso_ionize_recombine()" );

	/* get total charge transfer ionization rate if this is hydrogen itself,
	 * routine also does CT for helium but all models must have hydrogen, so
	 * he is done at this stage */
	if( ipISO==ipH_LIKE && nelem == ipHYDROGEN )
		ChargeTransferUpdate();

	/* find recombination and ionization elements, will use to get simple estimate
	 * of ionization ratio below */
	/* >>chng 06 jul 20, level should go to numLevels_local instead of numLevels_max */
	for( level=ipH1s; level< iso.numLevels_local[ipISO][nelem]; ++level)
	{
		/* all process moving level to continuum, units s-1 */
		iso.RateLevel2Cont[ipISO][nelem][level] = iso.gamnc[ipISO][nelem][level] + 
		  iso.ColIoniz[ipISO][nelem][level]* dense.EdenHCorr + 
		  /* >>chng 04 nov 30, also turn off cr ionization with atom h-like collisions off */
		  /* >>chng 05 aug 17, do not turn off cosmic rays with no collisional ionization 
		   * simply leave out crs if this is desired - this allows leiden hacks to work */
		  secondaries.csupra[nelem][nelem-ipISO]/* *iso.lgColl_ionize[ipISO]*/;

		/* all processes from continuum to level n, units s-1 */
		iso.RateCont2Level[ipISO][nelem][level] = (
			/* radiative recombination */
			iso.RadRecomb[ipISO][nelem][level][ipRecRad]*
			iso.RadRecomb[ipISO][nelem][level][ipRecNetEsc] + 

			/* induced recombination */
			iso.RecomInducRate[ipISO][nelem][level]*iso.PopLTE[ipISO][nelem][level] + 

			/* collisional or three body recombination */
			/* PopLTE(level,nelem) is only LTE pop when mult by n_e n_H */
			iso.ColIoniz[ipISO][nelem][level]*dense.EdenHCorr*iso.PopLTE[ipISO][nelem][level]
			) * dense.eden;
	}

	/* all following go into or out of ground state */
	level = 0;

	/* grain charge transfer recombination and ionization, 
	 * assume goes into and comes from ground state */
	/* first grain surface losses - both to lower and higher stages of ionization */
	sum = 0.;
	for(ion=0; ion<nelem+1; ++ion )
		if( ion!=nelem-ipISO )
			sum += gv.GrainChTrRate[nelem][nelem-ipISO][ion];
	/*>>chng 04 sep 08, replaced grain ct rate from old sum with current sum */
	iso.RateLevel2Cont[ipISO][nelem][level] += sum;

	/* >>chng 04 apr 10, add formally correct grain surface ionization and recombination - 
	 * this HAD NOT been included for H-like species */
	/* >>chng 04 sep 08, use explicit term from higher stage of ionization - note that added
	 * GrainCreat[nelem][nelem-ipISO] for case of atomic He incorrectly added two-stage recombination
	 * to existing matrix logic */
	/*iso.RateCont2Level[ipISO][nelem][level] += ionbal.GrainCreat[nelem][nelem-ipISO];*/
	iso.RateCont2Level[ipISO][nelem][level] += gv.GrainChTrRate[nelem][nelem+1-ipISO][nelem-ipISO];

	/* now charge transfer - all into/from ground, two cases, H and not H */
	if( nelem==ipHYDROGEN )
	{
		/* charge transfer, hydrogen onto everything else */
		/* charge exchange ionization */
		iso.RateLevel2Cont[ipISO][nelem][ipH1s] += atmdat.HCharExcIonTotal;
		/* charge exchange recombination */
		iso.RateCont2Level[ipISO][nelem][ipH1s] += atmdat.HCharExcRecTotal;
	}
	else if( nelem==ipHELIUM && ipISO==ipHE_LIKE )
	{
		/* add CO here, only for he itself, 
		 * destruction of CO but recombination for he */
		/* >>chng 04 feb 14, update date to 1.4ee-9 from 1.1e-9,
		 * make var so auto keep parallel with co.c */

		/* The following terms in co.c destroy He+ and thereby need to be included 
		 * here.  Also included in the sum of co.hep_destroy are the contributions
		 * to the recombination due to hmole_step.*/

		co.hep_destroy = (float) CO_sink_rate("He+") +
			hmi.H2_total*hmi.rheph2hpheh +      hmi.heph2heh2p*hmi.H2_total;
			
		iso.RateCont2Level[ipISO][nelem][ipHe1s1S] += co.hep_destroy;
		/*fprintf(ioQQQ,"DEBUG co hep destroy %e\n",co.hep_destroy );*/

		/* this is ioniz of He due to ct with all other gas constituents */
		iso.RateLevel2Cont[ipISO][nelem][ipHe1s1S] += atmdat.HeCharExcIonTotal;
		/* this is recom of He due to ct with all other gas constituents */
		iso.RateCont2Level[ipISO][nelem][ipHe1s1S] += atmdat.HeCharExcRecTotal;
	}
	else
	{
		iso.RateCont2Level[ipISO][nelem][level] +=
			atmdat.HeCharExcRecTo[nelem][nelem-ipISO]*iso.Pop2Ion[ipHE_LIKE][ipHELIUM][ipHe1s1S]*dense.xIonDense[ipHELIUM][1] + 
			atmdat.HCharExcRecTo[nelem][nelem-ipISO]*iso.Pop2Ion[ipH_LIKE][ipHYDROGEN][ipH1s]*dense.xIonDense[ipHYDROGEN][1];

		iso.RateLevel2Cont[ipISO][nelem][level] +=
			atmdat.HeCharExcIonOf[nelem][nelem-ipISO]*dense.xIonDense[ipHELIUM][1] +
			atmdat.HCharExcIonOf[nelem][nelem-ipISO]*dense.xIonDense[ipHYDROGEN][1];
	}


	/* now create sums of recombination and ionization rates for ISO species */
	ionbal.RateRecomTot[nelem][nelem-ipISO] = 0.;
	ionbal.RR_rate_coef_used[nelem][nelem-ipISO] = 0.;
	Recom3Body = 0.;
	/* >>chng 06 jul 20, level should go to numLevels_local instead of numLevels_max */
	for( level=ipH1s; level< iso.numLevels_local[ipISO][nelem]; ++level)
	{

		/* units of ionbal.RateRecomTot are s-1,
		 * equivalent ionization term is done after level populations are known */
		ionbal.RateRecomTot[nelem][nelem-ipISO] += iso.RateCont2Level[ipISO][nelem][level];

		/* just the radiative recombination rate coef, cm3 s-1 */
		ionbal.RR_rate_coef_used[nelem][nelem-ipISO] += iso.RadRecomb[ipISO][nelem][level][ipRecRad]*
			iso.RadRecomb[ipISO][nelem][level][ipRecNetEsc];
		/*if( nelem==ipHELIUM && ipISO==ipH_LIKE )
			fprintf(ioQQQ,"DEBUG rad rec %li %.4e %.4e\n",
				level,
				iso.RadRecomb[ipISO][nelem][level][ipRecRad],
				iso.RadRecomb[ipISO][nelem][level][ipRecNetEsc]);*/

		/* >>chng 05 jul 11, from > to >=, some very opt thick sims did block escape to zero */
		ASSERT( ionbal.RR_rate_coef_used[nelem][nelem-ipISO]>= 0. );

		/* this is three-body recombination rate coef by itself - 
		 * need factor of eden to become rate */
		Recom3Body += iso.ColIoniz[ipISO][nelem][level]*dense.EdenHCorr*iso.PopLTE[ipISO][nelem][level];
	}
	/*if( nelem==ipHELIUM && ipISO==ipH_LIKE )
		fprintf(ioQQQ,"DEBUG rad rec %.4e\n",ionbal.RR_rate_coef_used[nelem][nelem-ipISO]);*/

	/* fraction of total recombs due to three body - when most are due to this
	 * small changes in temperature can produce large changes in rec coef,
	 * and so in ionization */
	iso.RecomCollisFrac[ipISO][nelem] = (float)(Recom3Body* dense.eden / ionbal.RateRecomTot[nelem][nelem-ipISO] );

	/* get simple estimate of level of ionization */
	if( ionbal.RateRecomTot[nelem][nelem-ipISO] > 0. )
	{
#		if 0
		/* >>chng 06 jul 23, this was a temporary hack to force the lot T solver, which forces the
		 * atom to agree with iso.xIonSimple, to synch up with full solln from matrix - this caused
		 * many problems since using ionbal.RateIonizTot includes excited level ionization processes
		 * so matrix soln was called rather than simple.  this cause neg pops in ~half a dozen sims */
		if( ionbal.RateIonizTot[nelem][nelem-ipISO] >= 0. )
		{
			/* >>chng 06 jul 21, use effective rate, which includes ioniz from excited states, if defined */
			iso.xIonSimple[ipISO][nelem] = ionbal.RateIonizTot[nelem][nelem-ipISO]/ionbal.RateRecomTot[nelem][nelem-ipISO];
		}
		else
#		endif
		{
			iso.xIonSimple[ipISO][nelem] = iso.RateLevel2Cont[ipISO][nelem][ipH1s]/ionbal.RateRecomTot[nelem][nelem-ipISO];
		}
		/*fprintf(ioQQQ,"DEBUG iso ion rec %li %li ion %.2e ioniztot %.2e \n" ,
			ipISO, nelem ,
			iso.RateLevel2Cont[ipISO][nelem][ipH1s] ,
			ionbal.RateIonizTot[nelem][nelem-ipISO] );*/
	}
	else
	{
		iso.xIonSimple[ipISO][nelem] = 0.;
	}

	DEBUG_EXIT( "iso_ionize_recombine()" );

	return;
}
/*lint +e778 constant expression evaluates to zero - in HMRATE macro */
