/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
#include "cddefines.h"
#include "path.h"

/**\file path.cpp
 *
 * Cloudy needs to read a large number of data files when it starts.
 * The location of these data on your system is specified by 
 *
 * #define LOCAL_DATA_PATH "XXXX" 
 *
 * below, so the code is set up so that data will be automatically
 * located.  "XXXX" gives the location of the data files on your computer.
 *
 * This can also be set by using an argument to the compiler.  That is
 * discussed on the web site at 
 * https://cloud9.pa.uky.edu/trac/cloudy/wiki/EditPath
 *
 * The data path can also be set by the CLOUDY_DATA_PATH shell
 * variable, but it is generally better to provide the
 * permanent location at compile time by editing LOCAL_DATA_PATH below.
 */

/* The value below will be superseded if
 * specified by a compiler argument */
#ifndef LOCAL_DATA_PATH 

/* 
 * Specify path to local data here.
 *
 * This is for Unix variants.  You specify the data directory using the usual
   Unix file name format, e.g.: */
#define LOCAL_DATA_PATH "/u/home1/gary/cloudy/current/data/"

/* The following shows the format for Windows.  The filename is the
 * value between the "" -- the '\' character in Windows pathnames
 * needs to be included twice for each '\' in the file name.*/
// #define LOCAL_DATA_PATH "c:\\projects\\cloudy\\trunk\\data\\" 

/* The // makes the above line of code a comment
 * if you want to use the Windows version remove the // from the line above and 
 * place it place it in front of the unix path 9 lines above this line */

#endif

/* That should be all you need to change! */

/* FILENAME_PATH_LENGTH is the size of the string that holds the path.
	 Increase the size of this variable in the header file cddefines.h
	 if a larger string is needed to hold the path on your system.  */

char chDataPath[FILENAME_PATH_LENGTH]=LOCAL_DATA_PATH;
/* true says that the path has been set - do not change this */
bool lgDataPathSet=true;
