/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*state_get_put get or save state - called by cloudy - job is either "get" or "put" */
/*state_do - worker to actually get or put the structure -
 * called by state_get_put below */
#include "cddefines.h"
#include "taulines.h"
#include "iso.h"
#include "struc.h"
#include "mole.h"
#include "rfield.h"
#include "iterations.h"
#include "h2.h"
#include "dense.h"
#include "opacity.h"
#include "atomfeii.h"
#include "state.h"

static bool lgGet;
static FILE *ioSTATE;

/*state_do - worker to actually get or put the structure -
 * called by state_get_put below */
static void state_do( void *pnt , size_t sizeof_pnt )
{
	size_t n;
	double sanity = 1.,
		chk_sanity;
	size_t sizeof_sanity =sizeof( sanity );

	DEBUG_ENTRY( "state_do()" );

	/* nothing to do if this is not positive */
	if( sizeof_pnt == 0 )
		return;

	if( lgGet )
	{
		/* get option - read in data */
		if( (n=fread( pnt , 1 , sizeof_pnt , ioSTATE )) - sizeof_pnt )
		{
			fprintf( ioQQQ, " state_do failed reading state file, wanted %lu got %lu\n",
				(unsigned long)sizeof_pnt ,
				(unsigned long)n);
			puts( "[Stop in state_do]" );
			cdEXIT(EXIT_FAILURE);
		}
		/* perform sanity check */
		if( (n=fread( &chk_sanity , 1 , sizeof_sanity , ioSTATE )) - sizeof_sanity )
		{
			fprintf( ioQQQ, " state_do failed reading sanity par of state file, wanted %lu got %lu\n",
				(unsigned long)sizeof_sanity ,
				(unsigned long)n);
			puts( "[Stop in state_do]" );
			cdEXIT(EXIT_FAILURE);
		}
		/*lint -e777 check float equal */
		if( sanity != chk_sanity )
		/*lint +e777 check float equal */
		{
			fprintf( ioQQQ, " state_do sanity fails in state file, wanted %g got %g\n",
				sanity ,
				chk_sanity);
			puts( "[Stop in state_do]" );
			cdEXIT(EXIT_FAILURE);
		}
	}
	else
	{
		/* put option - write out data */
		fwrite( pnt , 1 , sizeof_pnt , ioSTATE );
		/* write sanity check */
		fwrite( &sanity , 1 , sizeof_sanity , ioSTATE );
	}

	DEBUG_EXIT( "state_do()" );

	return;
}

/*state_get_put get or save state - called by cloudy - job is either "get" or "put" */
void state_get_put( const char chJob[] )
{
	long int ipISO , nelem , ipHi, i ,
		iElecHi , iVibHi , iRotHi , iElecLo , iVibLo,
		n , ion;
	static FILE *ioGET_STATE , 
		*ioPUT_STATE;

	DEBUG_ENTRY( "state_get_put()" );

	if( (strcmp( chJob , "get" ) == 0) )
	{
		lgGet = true;
		if( (ioGET_STATE = fopen( state.chGetFilename , "rb" ))==NULL )
		{
			fprintf( ioQQQ, " state_get_put cannot open %s for reading.\n Sorry.", state.chGetFilename );
			puts( "[Stop in state_get_put]" );
			cdEXIT(EXIT_FAILURE);
		}
		ioSTATE = ioGET_STATE;
	}
	else if( (strcmp( chJob , "put" ) == 0) )
	{
		char chFilename[INPUT_LINE_LENGTH];
		if( !state.lgPutAll && iteration <= iterations.itermx )
		{

			DEBUG_EXIT( "state_get_put()" );
			/* not last iteration and do not want to save state for all iters, so simply quit */
			return;
		}

		/* get base of file name */
		strcpy( chFilename , state.chPutFilename );
		/* append iteration number if ALL keyword appears */
		if( state.lgPutAll )
		{
			char chIteration[INPUT_LINE_LENGTH];
			sprintf( chIteration , "_%li", iteration );
			strcat( chFilename , chIteration );
		}
		if( (ioPUT_STATE = fopen( chFilename , "wb" ))==NULL )
		{
			fprintf( ioQQQ, " state_get_put cannot open %s for writing.\n Sorry.", state.chPutFilename );
			puts( "[Stop in state_get_put]" );
			cdEXIT(EXIT_FAILURE);
		}
		ioSTATE = ioPUT_STATE;
	}
	else
		TotalInsanity();
	
	if( state.lgState_print )
		fprintf(ioQQQ," Print state quantities, start iso seq \n");
	/* actually do the read / write */
	for( ipISO=ipH_LIKE; ipISO<NISO; ++ipISO )
	{
		/* loop over all iso-electronic sequences */
		for( nelem=ipISO; nelem<LIMELM; ++nelem )
		{
			if( nelem < 2 || dense.lgElmtOn[nelem] )
			{
				/* arrays are dim'd iso.numLevels_max[ipH_LIKE][nelem]+1 */
				for( ipHi=1; ipHi < iso.numLevels_max[ipISO][nelem]; ++ipHi )
				{
					state_do( EmisLines[ipISO][nelem][ipHi] , 
						sizeof(EmLine)*(size_t)ipHi );
					if( state.lgState_print )
					{
						fprintf(ioQQQ," start ISO ipISO= %li, nelem= %li, ipHi %li \n", 
							ipISO , 
							nelem ,
							ipHi);
						for( n=0; n< ipHi; ++n )
						{
							fprintf(ioQQQ," ISO %li %li %li %li %.4e %.4e \n",
								ipISO , nelem , ipHi , n , 
								EmisLines[ipISO][nelem][ipHi][n].TauIn ,
								EmisLines[ipISO][nelem][ipHi][n].TauTot );
						}
						fprintf(ioQQQ," end ISO ipISO\n");
					}
				}

				if( state.lgState_print )
				{
					fprintf(ioQQQ," start Ext ipISO= %li, nelem= %li, got %li \n", 
						ipISO , 
						nelem ,
						iso.nLyman_malloc[ipISO] );
				}
				state_do( iso.ExtraLymanLines[ipISO][nelem] , 
					sizeof(EmLine )*(size_t)iso.nLyman_malloc[ipISO] );
				if( state.lgState_print )
				{
					for( n=2; n< iso.nLyman_malloc[ipISO]; ++n )
					{
						fprintf(ioQQQ," Ext %li %li %li %.4e %.4e \n",
							ipISO , nelem , n , 
							iso.ExtraLymanLines[ipISO][nelem][n].TauIn ,
							iso.ExtraLymanLines[ipISO][nelem][n].TauTot );
					}
					fprintf(ioQQQ," end Ext ipISO\n");
				}
			}
		}
	}

	state_do( TauLines    , (size_t)(nLevel1+1)*sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< (nLevel1+1); ++n )
		{
			fprintf(ioQQQ," Taulines %li %.4e %.4e \n",
				n , 
				TauLines[n].TauIn ,
				TauLines[n].TauTot );
		}
	}

	state_do( TauLine2    , (size_t)nWindLine  *sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< nWindLine; ++n )
		{
			fprintf(ioQQQ," TauLine2 %li %.4e %.4e \n",
				n , 
				TauLine2[n].TauIn ,
				TauLine2[n].TauTot );
		}
	}

	state_do( UTALines    , (size_t)nUTA   *sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< nUTA; ++n )
		{
			fprintf(ioQQQ," UTALines %li %.4e %.4e \n",
				n , 
				UTALines[n].TauIn ,
				UTALines[n].TauTot );
		}
	}

	state_do( HFLines     , (size_t)nHFLines   *sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< nHFLines; ++n )
		{
			fprintf(ioQQQ," HFLines %li %.4e %.4e \n",
				n , 
				HFLines[n].TauIn ,
				HFLines[n].TauTot );
		}
	}

	state_do( C12O16Rotate, (size_t)nCORotate  *sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< nCORotate; ++n )
		{
			fprintf(ioQQQ," C12O16Rotate %li %.4e %.4e \n",
				n , 
				C12O16Rotate[n].TauIn ,
				C12O16Rotate[n].TauTot );
		}
	}

	state_do( C13O16Rotate, (size_t)nCORotate  *sizeof(EmLine ) );
	if( state.lgState_print )
	{
		for( n=0; n< nCORotate; ++n )
		{
			fprintf(ioQQQ," C13O16Rotate %li %.4e %.4e \n",
				n , 
				C13O16Rotate[n].TauIn ,
				C13O16Rotate[n].TauTot );
		}
	}

	for( ipHi=1; ipHi < FeII.nFeIILevel; ++ipHi )
	{
		state_do( Fe2LevN[ipHi] , (size_t)ipHi*sizeof(EmLine) );
		if( state.lgState_print )
		{
			for( n=0; n< ipHi; ++n )
			{
				fprintf(ioQQQ," Fe2LevN %li %li %.4e %.4e \n",
					ipHi , n , 
					Fe2LevN[ipHi][n].TauIn ,
					Fe2LevN[ipHi][n].TauTot );
			}
		}
	}
	for( i=0; i<2; ++i )
	{
		state_do( opac.TauAbsGeo[i]  , (size_t)rfield.nupper*sizeof(float) );
		if( state.lgState_print )
		{
			for( n=0; n< rfield.nupper; ++n )
			{
				fprintf(ioQQQ," TauAbsGeo %li %li %.4e \n",
					i , n , 
					opac.TauAbsGeo[i][n] );
			}
		}

		state_do( opac.TauScatGeo[i] , (size_t)rfield.nupper*sizeof(float) );
		if( state.lgState_print )
		{
			for( n=0; n< rfield.nupper; ++n )
			{
				fprintf(ioQQQ," TauScatGeo %li %li %.4e \n",
					i , n , 
					opac.TauAbsGeo[i][n] );
			}
		}

		state_do( opac.TauTotalGeo[i], (size_t)rfield.nupper*sizeof(float) );
		if( state.lgState_print )
		{
			for( n=0; n< rfield.nupper; ++n )
			{
				fprintf(ioQQQ," TauTotalGeo %li %li %.4e \n",
					i , n , 
					opac.TauAbsGeo[i][n] );
			}
		}

	}

	/* the large H2 molecule, only if on */
	if( h2.lgH2ON )
	{
		for( iElecHi=0; iElecHi<mole.n_h2_elec_states; ++iElecHi )
		{
			for( iVibHi=0; iVibHi<=h2.nVib_hi[iElecHi]; ++iVibHi )
			{
				for( iRotHi=h2.Jlowest[iElecHi]; iRotHi<=h2.nRot_hi[iElecHi][iVibHi]; ++iRotHi )
				{
					/* NB - X is the only lower level considered here, since we are only 
					 * concerned with excited electronic levels as a photodissociation process
					 * code exists to relax this assumption - simply change following to iElecHi */
					long int lim_elec_lo = 0;
					for( iElecLo=0; iElecLo<=lim_elec_lo; ++iElecLo )
					{
						/* want to include all vib states in lower level if different elec level,
						 * but only lower vib levels if same elec level */
						long int nv = h2.nVib_hi[iElecLo];
						if( iElecLo==iElecHi )
							nv = iVibHi;
						for( iVibLo=0; iVibLo<=nv; ++iVibLo )
						{
							long nr = h2.nRot_hi[iElecLo][iVibLo];
							if( iElecLo==iElecHi && iVibHi==iVibLo )
								/* max because cannot malloc 0 bytes */
								nr = MAX2(1,iRotHi-1);
							state_do( H2Lines[iElecHi][iVibHi][iRotHi][iElecLo][iVibLo] ,
								(size_t)(nr+1)*sizeof(EmLine) );
						}
					}
				}
			}
		}
	}

	/* the struc variables */
	state_do( &struc.nzlim, sizeof(struc.nzlim ) );
	state_do( &struc.dr_ionfrac_limit, sizeof(struc.dr_ionfrac_limit ) );
	state_do( &struc.nzone, sizeof(struc.nzone ) );

	state_do( struc.testr, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.volstr, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.drad_x_fillfac, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.drad, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.histr, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.hiistr, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.ednstr, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.o3str, (size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.pressure,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.GasPressure ,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.pres_radiation_lines_curr ,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.hden ,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.DenParticles ,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.DenMass,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.depth,(size_t)(struc.nzlim)*sizeof(float ) );
	state_do( struc.xLyman_depth , (size_t)(struc.nzlim)*sizeof(float ) );

	state_do( struc.coolstr,(size_t)(struc.nzlim)*sizeof(double ) );
	state_do( struc.heatstr , (size_t)(struc.nzlim)*sizeof(double ) );

	for( nelem=ipHYDROGEN; nelem<(LIMELM+3); ++nelem )
	{
		for( ion=0; ion<(LIMELM+1); ++ion )
		{
			state_do( struc.xIonDense[nelem][ion] , (size_t)(struc.nzlim)*sizeof(float ) );
		}
	}

	/* the hydrogen molecules */
	/*float *Molec[N_H_MOLEC];*/
	for( n=0; n<N_H_MOLEC; ++n )
	{
		state_do( struc.H2_molec[n] , (size_t)(struc.nzlim)*sizeof(float ) );
	}
	for( n=0; n<mole.num_comole_calc; ++n )
	{
		state_do( struc.CO_molec[n] , (size_t)(struc.nzlim)*sizeof(float ) );
	}

	for( nelem=ipHYDROGEN; nelem<LIMELM; ++nelem )
	{
		state_do( struc.gas_phase[nelem] , (size_t)(struc.nzlim)*sizeof(float ) );
	}

	/*fprintf(ioQQQ,"DEBUG done\n");*/

	/* say that outer optical depths are defined */
	opac.lgTauOutOn = true;

	/* close the file */
	fclose( ioSTATE );

	DEBUG_EXIT( "state_get_put()" );

	return;
}

